/*
 *************************************************************************
 *
 * serdisp_tools.c
 * some common routines
 *
 *************************************************************************
 *
 * copyright (C) 2003-2007  wolfgang astleitner
 * email     mrwastl@users.sourceforge.net
 *
 *************************************************************************
 * This program is free software; you can redistribute it and/or modify   
 * it under the terms of the GNU General Public License as published by   
 * the Free Software Foundation; either version 2 of the License, or (at  
 * your option) any later version.                                        
 *                                                                        
 * This program is distributed in the hope that it will be useful, but    
 * WITHOUT ANY WARRANTY; without even the implied warranty of             
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU      
 * General Public License for more details.                               
 *                                                                        
 * You should have received a copy of the GNU General Public License      
 * along with this program; if not, write to the Free Software            
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA              
 * 02111-1307, USA.  Or, point your browser to                            
 * http://www.gnu.org/copyleft/gpl.html                                   
 *************************************************************************
 */

#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <strings.h>

#if defined(__NetBSD__) || defined(__FreeBSD__) || defined(__OpenBSD__)
 #include <sys/time.h>
#else
 #include <sys/resource.h>
#endif

#include "serdisplib/serdisp_tools.h"
#include "serdisplib/serdisp_messages.h"

/* to be sure to undef malloc vs. rpl_malloc hacks done by configure */
#undef malloc

/* table containing  byte = swapped(byte) */
/* taken from graphlcd/drivers/fct.c */
static const int bit_reverse_table[256] = {
0x00, 0x80, 0x40, 0xc0, 0x20, 0xa0, 0x60, 0xe0,
0x10, 0x90, 0x50, 0xd0, 0x30, 0xb0, 0x70, 0xf0,
0x08, 0x88, 0x48, 0xc8, 0x28, 0xa8, 0x68, 0xe8,
0x18, 0x98, 0x58, 0xd8, 0x38, 0xb8, 0x78, 0xf8,
0x04, 0x84, 0x44, 0xc4, 0x24, 0xa4, 0x64, 0xe4,
0x14, 0x94, 0x54, 0xd4, 0x34, 0xb4, 0x74, 0xf4,
0x0c, 0x8c, 0x4c, 0xcc, 0x2c, 0xac, 0x6c, 0xec,
0x1c, 0x9c, 0x5c, 0xdc, 0x3c, 0xbc, 0x7c, 0xfc,
0x02, 0x82, 0x42, 0xc2, 0x22, 0xa2, 0x62, 0xe2,
0x12, 0x92, 0x52, 0xd2, 0x32, 0xb2, 0x72, 0xf2,
0x0a, 0x8a, 0x4a, 0xca, 0x2a, 0xaa, 0x6a, 0xea,
0x1a, 0x9a, 0x5a, 0xda, 0x3a, 0xba, 0x7a, 0xfa,
0x06, 0x86, 0x46, 0xc6, 0x26, 0xa6, 0x66, 0xe6,
0x16, 0x96, 0x56, 0xd6, 0x36, 0xb6, 0x76, 0xf6,
0x0e, 0x8e, 0x4e, 0xce, 0x2e, 0xae, 0x6e, 0xee,
0x1e, 0x9e, 0x5e, 0xde, 0x3e, 0xbe, 0x7e, 0xfe,
0x01, 0x81, 0x41, 0xc1, 0x21, 0xa1, 0x61, 0xe1,
0x11, 0x91, 0x51, 0xd1, 0x31, 0xb1, 0x71, 0xf1,
0x09, 0x89, 0x49, 0xc9, 0x29, 0xa9, 0x69, 0xe9,
0x19, 0x99, 0x59, 0xd9, 0x39, 0xb9, 0x79, 0xf9,
0x05, 0x85, 0x45, 0xc5, 0x25, 0xa5, 0x65, 0xe5,
0x15, 0x95, 0x55, 0xd5, 0x35, 0xb5, 0x75, 0xf5,
0x0d, 0x8d, 0x4d, 0xcd, 0x2d, 0xad, 0x6d, 0xed,
0x1d, 0x9d, 0x5d, 0xdd, 0x3d, 0xbd, 0x7d, 0xfd,
0x03, 0x83, 0x43, 0xc3, 0x23, 0xa3, 0x63, 0xe3,
0x13, 0x93, 0x53, 0xd3, 0x33, 0xb3, 0x73, 0xf3,
0x0b, 0x8b, 0x4b, 0xcb, 0x2b, 0xab, 0x6b, 0xeb,
0x1b, 0x9b, 0x5b, 0xdb, 0x3b, 0xbb, 0x7b, 0xfb,
0x07, 0x87, 0x47, 0xc7, 0x27, 0xa7, 0x67, 0xe7,
0x17, 0x97, 0x57, 0xd7, 0x37, 0xb7, 0x77, 0xf7,
0x0f, 0x8f, 0x4f, 0xcf, 0x2f, 0xaf, 0x6f, 0xef,
0x1f, 0x9f, 0x5f, 0xdf, 0x3f, 0xbf, 0x7f, 0xff
};


/* table containing bitmasks needed for serdisp_get/setpixel() */
static const byte pixel_bitmask_table[] = {
   0,  /* depth 0: non existant */
0x01,  /* depth 1: 0000 0001 */
0x03,  /* depth 2: 0000 0011 */
   0,  /* depth 3: not supported */
0x0F,  /* depth 4: 0000 1111 */
   0,  /* depth 5: not supported */
   0,  /* depth 6: not supported */
   0,  /* depth 7: not supported */
   0   /* depth 8: not supported */
};



/* *********************************
   char* sdtools_strncpy(dest, src, n);
   *********************************
   strncpy with workarounds avoiding strncpy-bugs found on solaris10 and/or glibc-linked programs
   *********************************
   dest     ... destination string
   src      ... source string
   n        ... amount of characters to be copied
*/
char* sdtools_strncpy(char *dest, const char *src, size_t n) {
#if defined(__sun__)
  return (char*) strlcpy(dest, src, n+1);
#else
  char* rv = strncpy(dest, src, n);
  dest[ (strlen(src) < n) ? strlen(src) : n ] = '\0';
  return rv;
#endif
}



/* *********************************
   void* sdtools_malloc(size);
   *********************************
   0 byte save malloc
   *********************************
   size     ... amount of bytes to allocate
*/
void *sdtools_malloc(size_t size) {
  if (!size) return (void*)0;
  return (void*)malloc(size);
}



/* *********************************
   void sdtools_generic_rotate(dd)
   *********************************
   rotates the display buffer (and updates the display)
   *********************************
   dd     ... display descriptor
*/
void sdtools_generic_rotate(serdisp_t* dd) {
  int x, y, w, h, temp;

  w = serdisp_getwidth(dd);
  h = serdisp_getheight(dd);

  for (y = 0; y < h; y++)
    for (x = 0; x < (w+1)/2; x++) {
      temp = serdisp_getpixel(dd, x, y);
      serdisp_setpixel(dd, x, y,  serdisp_getpixel(dd, w-1-x, h-1-y));
      serdisp_setpixel(dd, w-1-x, h-1-y, temp);
    }
  serdisp_update(dd);      
}


/* *********************************
   void sdtools_generic_setpixel(dd, x, y, colour)
   *********************************
   changes a pixel into the display buffer
   *********************************
   dd     ... display descriptor
   x      ... x-position
   y      ... y-position
   colour ... monochrome: 0: clear (white), <>0: set (black); else: grey value (supported depths: 1, 2, 4)
*/
void sdtools_generic_setpixel (serdisp_t* dd, int x, int y, long colour) {
  int x_i = 0, y_i = 0;

  if (dd->curr_rotate <= 1) {
    if (x >= dd->width || y >= dd->height || x < 0 || y < 0)
      return;
  } else {
    if (x >= dd->height || y >= dd->width || x < 0 || y < 0)
      return;
  }

  switch (dd->curr_rotate) {
    case 0:  /* 0 degrees */
      x_i = dd->xreloctab[x];
      y_i = dd->yreloctab[y];
      break;
    case 1:  /* 180 degrees */
      x_i = dd->xreloctab[dd->width  - 1 - x];
      y_i = dd->yreloctab[dd->height - 1 - y];
      break;
    case 2:  /* 90 degrees */
      x_i = dd->xreloctab[y];
      y_i = dd->yreloctab[dd->height  - 1 - x];
      break;
    case 3:  /* 270 degrees */
      x_i = dd->xreloctab[dd->width - 1 - y];
      y_i = dd->yreloctab[x];
      break;
  }

  if (dd->depth < 8) {
    int page, pagediv = 8 / dd->depth;
    byte value_orig, value, mask;
    byte coltrans;    /* mask out all inapplicable bits */
    page = y_i / pagediv;

    value_orig = dd->scrbuf[ (dd->width + dd->xcolgaps) * page  +  x_i];

    mask = pixel_bitmask_table[dd->depth] << ((y_i % pagediv)*dd->depth);
    coltrans = (byte)colour & pixel_bitmask_table[dd->depth];
    value = ((0xFF ^ mask) & value_orig) | (coltrans << (((y_i % pagediv))*dd->depth));

    if (value_orig != value) {
      dd->scrbuf[ (dd->width + dd->xcolgaps) * page  +  x_i] = value;

      if (dd->scrbuf_chg) {
        int idx = x_i + (dd->width + dd->xcolgaps) *(page/8);

        /* catch out of bound errors that may occur because of a bug in the display driver */
        if (idx >= dd->scrbuf_chg_size) {
          sd_debug(0, "sdtools_generic_setpixel(): OUTOFBOUND: idx>=scrbuf_chg_size: %d >= %d   x/y: %d/%d  x_i/y_i: %d/%d", 
                      idx, dd->scrbuf_chg_size, x,y, x_i,y_i);
        } else {
          dd->scrbuf_chg[idx] |= (1 << (page%8));
        }
      }
    }
  } else {
    int changed = 0;
    int idx_2 = ((x_i  + y_i * dd->width) * (dd->depth << 1)) / 8;
    int idx = idx_2 >> 1;

    switch(dd->depth) {
      case 8:
        if (dd->scrbuf[idx] != colour) {
          dd->scrbuf[idx] = colour;
          changed = 1;
        }
        break;
      case 12:
        {
          byte r = (colour & 0x00000F00) >> 8;
          byte g = (colour & 0x000000F0) >> 4;
          byte b = (colour & 0x0000000F);
          if (idx_2 & 0x01) { /* RGB -> 2nd nipple of first byte, second byte */
            if ( (dd->scrbuf[idx] & 0x0F) != r) {
              dd->scrbuf[idx] =  (dd->scrbuf[idx] & 0xF0) + r;
              changed = 1;
            }
            if ( dd->scrbuf[idx+1] != ( (g<<4) + b) ) {
              dd->scrbuf[idx+1] =  (g<<4) + b;
              changed = 1;
            }
          } else {
            if ( dd->scrbuf[idx] != ( (r<<4) + g) ) {
              dd->scrbuf[idx] =  (r<<4) + g;
              changed = 1;
            }
            if ( (dd->scrbuf[idx+1] & 0xF0) != (b << 4)) {
              dd->scrbuf[idx+1] =  (dd->scrbuf[idx+1] & 0x0F) + (b << 4);
              changed = 1;
            }
          }
        }
        break;
      case 16:
        if ( dd->scrbuf[idx] != ((colour & 0xFF00) >> 8) ) {
          dd->scrbuf[idx] = ((colour & 0xFF00) >> 8);
          changed = 1;
        }
        if ( dd->scrbuf[idx+1] != (colour & 0xFF) ) {
          dd->scrbuf[idx+1] = (colour & 0x00FF);
          changed = 1;
        }
        break;
    }

    if (changed && dd->scrbuf_chg) {
      /* ceil( dd->width / 8) bytes per row. one byte == one change info (== one changed pixel) */
      idx = (x_i >> 3) + y_i * ((dd->width + 7 ) >> 3);

      /* catch out of bound errors that may occur because of a bug in the display driver */
      if (idx >= dd->scrbuf_chg_size) {
        sd_debug(0, "sdtools_generic_setpixel(): OUTOFBOUND: idx>=scrbuf_chg_size: %d >= %d   x/y: %d/%d  x_i/y_i: %d/%d", 
                    idx, dd->scrbuf_chg_size, x,y, x_i,y_i);
      } else
        dd->scrbuf_chg[ idx ] |=  (1 << (x_i % 8));
    }
  }
}


/* *********************************
   int sdtools_generic_getpixel(dd, x, y)
   *********************************
   changes a pixel in the display buffer. supported depths: 1, 2, 4
   *********************************
   dd     ... display descriptor
   x      ... x-position
   y      ... y-position
   *********************************
   returns the grey value (or hw dependend index) at (x/y)
*/
long sdtools_generic_getpixel (serdisp_t* dd, int x, int y) {
  int x_i = 0, y_i = 0;
  long value = 0L;

  if (dd->curr_rotate <= 1) {
    if (x >= dd->width || y >= dd->height || x < 0 || y < 0)
      return 0;
  } else {
    if (x >= dd->height || y >= dd->width || x < 0 || y < 0)
      return 0;
  }

  switch (dd->curr_rotate) {
    case 0:  /* 0 degrees */
      x_i = dd->xreloctab[x];
      y_i = dd->yreloctab[y];
      break;
    case 1:  /* 180 degrees */
      x_i = dd->xreloctab[dd->width  - 1 - x];
      y_i = dd->yreloctab[dd->height - 1 - y];
      break;
    case 2:  /* 90 degrees */
      x_i = dd->xreloctab[y];
      y_i = dd->yreloctab[dd->height  - 1 - x];
      break;
    case 3:  /* 270 degrees */
      x_i = dd->xreloctab[dd->width - 1 - y];
      y_i = dd->yreloctab[x];
      break;
  }

  if (dd->depth < 8) {
    int page, pagediv = 8 / dd->depth;
    byte mask;
    byte value_byte;

    page = y_i / pagediv;

    value_byte = dd->scrbuf[ (dd->width + dd->xcolgaps) * page  +  x_i];

    mask = pixel_bitmask_table[dd->depth] << ((y_i % pagediv)*dd->depth);

    value_byte &= mask;

    value = (long)(value_byte >> ((y_i % pagediv)*dd->depth));
  } else {
    int idx_2 = ((x_i  + y_i * dd->width) * (dd->depth << 1)) / 8;
    int idx = idx_2 >> 1;


    switch(dd->depth) {
      case 8:
        value = (long) dd->scrbuf[idx];
        break;
      case 12:
        {
          if (idx_2 & 0x01)
            value = (long)(((dd->scrbuf[idx] & 0x0F) << 8) +  dd->scrbuf[idx+1]);
          else
            value = (long)( (dd->scrbuf[idx] << 4) +  ((dd->scrbuf[idx+1] & 0xF0) >> 4));
        }
        break;
      case 16:
        value = (long)((dd->scrbuf[idx] << 8) + dd->scrbuf[idx+1]);
        break;
    }

  }

  return value;
}




/* *********************************
   void sdtools_generic_setpixel_greyhoriz(dd, x, y, colour)
   *********************************
   changes a pixel into the display buffer (horizontally organised w/o pages)
   *********************************
   dd     ... display descriptor
   x      ... x-position
   y      ... y-position
   colour ... monochrome: 0: clear (white), <>0: set (black); else: up to 256 greyscales (dependend on display)
*/
void sdtools_generic_setpixel_greyhoriz (serdisp_t* dd, int x, int y, long colour) {
  int x_i = 0, y_i = 0, col;
  byte value_orig, value, mask;
  int cols;
  int idx = -1;

  if (dd->curr_rotate <= 1) {
    if (x >= dd->width || y >= dd->height || x < 0 || y < 0)
      return;
  } else {
    if (x >= dd->height || y >= dd->width || x < 0 || y < 0)
      return;
  }

  switch (dd->curr_rotate) {
    case 0:  /* 0 degrees */
      x_i = dd->xreloctab[x];
      y_i = dd->yreloctab[y];
      break;
    case 1:  /* 180 degrees */
      x_i = dd->xreloctab[dd->width  - 1 - x];
      y_i = dd->yreloctab[dd->height - 1 - y];
      break;
    case 2:  /* 90 degrees */
      x_i = dd->xreloctab[y];
      y_i = dd->yreloctab[dd->height  - 1 - x];
      break;
    case 3:  /* 270 degrees */
      x_i = dd->xreloctab[dd->width - 1 - y];
      y_i = dd->yreloctab[x];
      break;
  }

  if (dd->depth == 1) {
    byte scrbuf_bits_used =
#if (SERDISP_VERSION_CODE > SERDISP_VERSION(1,97))
    dd->scrbuf_bits_used;
#else
    8;
#endif

    cols = (dd->width + dd->xcolgaps + scrbuf_bits_used -1) / scrbuf_bits_used;

    col = x_i / scrbuf_bits_used;

    value_orig = dd->scrbuf[ cols * y_i + col ];

    mask = 1 << (scrbuf_bits_used-1-(x_i % scrbuf_bits_used));

    value = (colour) ? (value_orig | mask) : (value_orig & (0xFF ^ mask));

    idx = ((cols + 7) / (8 / dd->depth) ) * y_i + (col / (8 / dd->depth) ) ;
  } else {
    int coldiv = 8 / dd->depth;
    int mask_shift = ( x_i % coldiv) * dd->depth;

    cols = (dd->width + dd->xcolgaps) / coldiv;

    col = x_i / coldiv;

    value_orig = dd->scrbuf [ cols * y_i + col ];

    mask = pixel_bitmask_table[dd->depth] << mask_shift;

    value = ((0xFF ^ mask) & value_orig) | ((byte)colour << mask_shift);

    idx = y_i * (cols >> 3) + (col >> 3);
/*fprintf(stderr, "value: 0x%02x   value_orig: 0x%02x    col=%d  mask=0x%02x\n", (byte)value, value_orig, col, mask);*/
  }

  if (value_orig != value) {
    dd->scrbuf[ cols * y_i + col ] = value;

    if (dd->scrbuf_chg) {
      /* catch out of bound errors that may occur because of a bug in the display driver */
      if (idx >= dd->scrbuf_chg_size) {
        sd_debug(0, "sdtools_generic_setpixel(): OUTOFBOUND: idx>=scrbuf_chg_size: %d >= %d   x/y: %d/%d  x_i/y_i: %d/%d", 
                    idx, dd->scrbuf_chg_size, x,y, x_i,y_i);
      } else {
        dd->scrbuf_chg[idx] |= (1 << (col % 8));
      }
    }
  }
}



/* *********************************
   int sdtools_generic_getpixel_greyhoriz(dd, x, y)
   *********************************
   changes a pixel in the display buffer
   *********************************
   dd     ... display descriptor
   x      ... x-position
   y      ... y-position
   *********************************
   returns the colour at (x/y)
*/
long sdtools_generic_getpixel_greyhoriz (serdisp_t* dd, int x, int y) {
  int x_i = 0, y_i = 0, col;
  byte value, mask;
  int cols;

  if (dd->curr_rotate <= 1) {
    if (x >= dd->width || y >= dd->height || x < 0 || y < 0)
      return 0;
  } else {
    if (x >= dd->height || y >= dd->width || x < 0 || y < 0)
      return 0;
  }

  switch (dd->curr_rotate) {
    case 0:  /* 0 degrees */
      x_i = dd->xreloctab[x];
      y_i = dd->yreloctab[y];
      break;
    case 1:  /* 180 degrees */
      x_i = dd->xreloctab[dd->width  - 1 - x];
      y_i = dd->yreloctab[dd->height - 1 - y];
      break;
    case 2:  /* 90 degrees */
      x_i = dd->xreloctab[y];
      y_i = dd->yreloctab[dd->height  - 1 - x];
      break;
    case 3:  /* 270 degrees */
      x_i = dd->xreloctab[dd->width - 1 - y];
      y_i = dd->yreloctab[x];
      break;
  }

  if (dd->depth == 1) {
    byte scrbuf_bits_used =
#if (SERDISP_VERSION_CODE > SERDISP_VERSION(1,97))
    dd->scrbuf_bits_used;
#else
    8;
#endif
    cols = (dd->width + dd->xcolgaps + scrbuf_bits_used -1) / scrbuf_bits_used;

    col = x_i / scrbuf_bits_used;

    value = dd->scrbuf[ cols * y_i + col ];

    mask = 1 << (scrbuf_bits_used-1-(x_i % scrbuf_bits_used));

    value = (value & mask) ? 1 : 0;
  } else {
    int coldiv = 8 / dd->depth;
    int mask_shift = ( x_i % coldiv) * dd->depth;

    cols = (dd->width + dd->xcolgaps) / coldiv;

    col = x_i / coldiv;

    value = dd->scrbuf [ cols * y_i + col ];

    mask = pixel_bitmask_table[dd->depth] << mask_shift;

    value = (value & mask) >> mask_shift;
  }
  return (value); 
}




/* *********************************
   void sdtools_nsleep (long ns)
   *********************************
     ns = 0:  no delay (do nothing)
     ns = 1:  do a very simple delay (one call of gettimeofday)
     ns > 1:   delay ns nanoseconds
   *********************************
   ns  ... amount of nanoseconds to delay 
   *********************************

   attention: i will NOT use scheduling and priority-calls as they 
              caused errors in MPEG-PES stream of vdr

*/
void sdtools_nsleep(long ns) {
  if (ns > 1) {
    /* source: picport.cc by Jaakko Hyvatti
       http://www.iki.fi/hyvatti
    */
    struct timeval tv1, tv2;
    volatile int i;

    gettimeofday(&tv1, 0);
    tv2.tv_sec = tv1.tv_sec;
    tv2.tv_usec = 0xffffffff & (tv1.tv_usec + 1 + (ns + 999) / 1000);
    if (tv2.tv_usec < tv1.tv_usec)
      tv2.tv_sec++;
    for (i = 0; i < 10000; i++) {
      gettimeofday(&tv1, 0);
      if (tv1.tv_sec > tv2.tv_sec || (tv1.tv_sec == tv2.tv_sec && tv1.tv_usec >= tv2.tv_usec))
        break;
    }
  } else if (ns == 1) {
    struct timeval tv1;
    gettimeofday(&tv1, 0);
  }
  /* if <= 0: ignore sdtools_nsleep */
}



/* *********************************
   byte sdtools_reversebits (byte b)
   *********************************
   swaps bits in a byte: bit7 becomes bit0, bit6 becomes bit1 and so on
   a predefined table containing all 256 dupels is used for faster execution 
   *********************************
   b  ... input byte
   *********************************
   returns swapped byte

   source: graphlcd/drivers/fct.c   
   http://www.powarman.de/ -> VDR Graphic LCD plugin
*/
byte sdtools_reversebits(byte b) {
  return bit_reverse_table[b];
}



/* *********************************
   char* sdtools_strlefttrim (str)
   *********************************
   returns a char-pointer to the first non-space/tabulator character in str
   if no leading spaces/tabulators: str is returned as is
   no extra memory is allocated
   *********************************
   str  ... string
   *********************************
   returns left-trimmed string
*/
char* sdtools_strlefttrim(const char* str) {
  int n = 0;
  
  int len = strlen(str);
  
  while ( n < len && (str[n] == ' ' || str[n] == '\t') )
    n++;
    
  return (char*) (&str[n]);
}


/* *********************************
   int sdtools_strtrimmedlen (str, len)
   *********************************
   returns length of str without any trailing spaces or tabulators
   (only len characters of str are used for this operation)
   if no trailing spaces/tabulators: len is returned
   
   examples:
     sdtools_strtimmedlen("test  ", 6) -> 4 will be returned
     sdtools_strtimmedlen("test  ,xyz", 6) -> 4 will be returned
     sdtools_strtimmedlen("test  ,xyz", 8) -> 8 will be returned
   
   no extra memory is allocated
   *********************************
   str  ... string
   len  ... how many characters to use from str
   *********************************
   returns length of str without trailing spaces or tabulators
*/
int sdtools_strtrimmedlen(const char* str, int len) {
  int n = len;
  
  while ( n > 0 && (str[n-1] == ' ' || str[n-1] == '\t') )
    n--;
    
  return n;  
}



/* *********************************
   int sdtools_isinelemlist (elemlist, str, strlen)
   *********************************
   tests if str is an element contained in comma-separated list elemlist. 
   leading and trailing spaces and tabs are ignored. commas are interpreted as separators and are not valid in elements!
   
   examples:
     sdtools_isinelemlist(" xyz , abc,cde ", "xyz")  ->  0 will be returned
     sdtools_isinelemlist(" xyz , abc,cde ", "abc")  ->  1 will be returned
     sdtools_isinelemlist(" xyz , abc,cde ", " cde") ->  2 will be returned (trailing spaces will be ignored)
     sdtools_isinelemlist(" xyz , abc,cde ", "ijk")  -> -1 will be returned
     sdtools_isinelemlist(" xyz , abc,cde ", "c,c")  -> -1 will be returned (',' is a separator but no valid character)
   *********************************
   elemlist ... comma separated element list
   str      ... element
   *********************************
   returns -1 if str is not found else its array index is returned
*/
int sdtools_isinelemlist(const char* elemlist, const char* str, int len) {
  char* lefttrimmedstr = sdtools_strlefttrim(str);
  int nstr;
  int cnt = 0;

  char* lefttrimmedelemlist = (char*)elemlist;
  int nelemlist;
  char* indexpos;

  nstr = (len == -1 ) ? strlen(lefttrimmedstr) : len - serdisp_ptrstrlen(lefttrimmedstr, str);
  nstr = sdtools_strtrimmedlen(lefttrimmedstr, nstr);
  
  do {
    lefttrimmedelemlist = sdtools_strlefttrim(lefttrimmedelemlist);
    
    if (strlen(lefttrimmedelemlist) == 0) return -1;
    
    indexpos = index(lefttrimmedelemlist, ',');
    
    /* this should never occur but to be sure: catch it and return 0 in such a case (invalid elemlist!) */
    if (indexpos == lefttrimmedelemlist)
      return -1;

    nelemlist = sdtools_strtrimmedlen(
                  lefttrimmedelemlist, 
                  ((indexpos) ? serdisp_ptrstrlen(indexpos, lefttrimmedelemlist) : strlen(lefttrimmedelemlist))
                );
    
    if (nelemlist == nstr && strncasecmp(lefttrimmedelemlist, lefttrimmedstr, nelemlist) == 0) {
      return cnt;
    } else {
      if (indexpos) {
        if (strlen(indexpos) <= 1) return -1;  /* trailing ',' in elemlist */

        lefttrimmedelemlist=&indexpos[1]; /* skip separator */
      }
    }
    
    cnt ++;
  } while (indexpos);
    
    
  return -1;  
}


/* *********************************
   char* sdtools_nextpattern (str, delim, len, border)
   *********************************
   get next pattern delimited with 'delim' or \0.
   leading spaces or tabs are ignored.
   
   len will contain the length of this pattern without any trainling spaces or tabs.
   for the first pattern, len has to be set to -1 and is set to the length of the found pattern
   
   border contains the maximum characters to take care of. all characters above border are ignored.
   border will be adapted when a new pattern was found   
   *********************************
   str     ... string
   delim   ... delimiter (eg: ',' or ';')
   &len    ... length of found pattern (w/o leading and trailing spaces or tabulators) (CALL BY REFERENCE)
   &border ... amount of characters that are valid for searching patterns              (CALL BY REFERENCE)
   *********************************
   returns pointer to the found pattern (with leading spaces and tabulators eliminated) or 0 if no more patterns are available
   
   
   examples:
     str = "elem1=val1, elem2= val2 ,  elem3=val3   ;  xyz=abc, def = ghi "
     border = 40;  (; and the following characters are ignored)
     len = -1;  (start with -1 to get first parameter)
     
     pattern = sdtools_nextpattern(str, ',', &len, &border);
     -> pattern will contain "elem1=........", len = 10; border = 40; so pattern => "elem1=val1"
     
     pattern = sdtools_nextpattern(str, ',', &len, &border);
     -> pattern will contain "elem2=........", len = 11; border = 22; so pattern => "elem2= val2"

     pattern = sdtools_nextpattern(str, ',', &len, &border);
     -> pattern will contain "elem3=........", len = 10; border = 13; so pattern => "elem3=val3"

     pattern = sdtools_nextpattern(str, ',', &len, &border);
     -> pattern will contain 0, because the border was exceeded

*/
char* sdtools_nextpattern(const char* str, char delim, int* len, int* border) {
  char* strstart = (char*) str;
  int n;
  char* idxpos;
  
  
  if (*len >= 0) {  /* *len < 0: return first parameter, else: not first parameter */
    idxpos = index(strstart, delim);
    if ( !idxpos || (((long)idxpos) >= (((long)strstart) + *border)) ) { 
      /* no more delim or found delim out of view => no more param */
      *len = -1; *border = 0;
      return 0;
    }
    strstart = idxpos+1;
    *border -=  serdisp_ptrstrlen(strstart, str);
    *len = *border;
  } else { /* first parameter */
    *len = *border;
  }

  n = 0;
  while ( n < *len && (strstart[n] == ' ' || strstart[n] == '\t') )
    n++;
    
  strstart = &(strstart[n]);
  *border -= n;
  *len -= n;
  
  idxpos = index(strstart, delim);

  if ( idxpos && ((long)idxpos < ((long)strstart + (*border))) ) {
    *len = serdisp_ptrstrlen(idxpos, strstart);
  } else {
    *len = *border;
  }

  while ( *len > 0 && (strstart[(*len)-1] == ' ' || strstart[(*len)-1] == '\t') )
    (*len)--;

  return strstart;  
}



/* *********************************
   int   sdtools_ismatching  (str1, len1, str2, len2) {
   *********************************
   compare two strings whether they're matching or not
   case-insensitive, leading spaces or tabs are ignored.
   
   *********************************
   str1    ... string 1
   len1    ... lenght of string 1  (everything after len1 will be ignored)
   str2    ... string 2
   len2    ... lenght of string 2  (everything after len1 will be ignored)
   *********************************
   returns 1 if strings are matching or 0 if not
*/
int sdtools_ismatching(const char* str1, int len1, const char* str2, int len2) {

  char* startptr1 = sdtools_strlefttrim(str1);
  char* startptr2 = sdtools_strlefttrim(str2);

  if (len1 == -1)
    len1 = strlen(str1);
  
  if (len2 == -1)
    len2 = strlen(str2);
  
  
  len1 -= serdisp_ptrstrlen(startptr1, str1);
  len2 -= serdisp_ptrstrlen(startptr2, str2);
  
  len1 = sdtools_strtrimmedlen(startptr1, len1);
  len2 = sdtools_strtrimmedlen(startptr2, len2);
  
  return (len1 == len2  && strncasecmp(startptr1, startptr2, len1) == 0);
}
